/*
 * Copyright 2010 ZXing authors
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import 'dart:typed_data';

import '../../common/bit_matrix.dart';
import '../../common/detector/math_utils.dart';
import '../../common/detector/white_rectangle_detector.dart';
import '../../common/grid_sampler.dart';
import '../../common/reedsolomon/generic_gf.dart';
import '../../common/reedsolomon/reed_solomon_decoder.dart';

import '../../not_found_exception.dart';
import '../../result_point.dart';
import '../aztec_detector_result.dart';

class Point {
  final int x;
  final int y;

  ResultPoint toResultPoint() {
    return ResultPoint(x.toDouble(), y.toDouble());
  }

  Point(this.x, this.y);

  int getX() {
    return x;
  }

  int getY() {
    return y;
  }

  @override
  String toString() {
    return "<$x $y>";
  }
}

/// Encapsulates logic that can detect an Aztec Code in an image, even if the Aztec Code
/// is rotated or skewed, or partially obscured.
///
/// @author David Olivier
/// @author Frank Yellin
class Detector {
  static const List<int> _EXPECTED_CORNER_BITS = [
    0xee0, // 07340  XXX .XX X.. ...
    0x1dc, // 00734  ... XXX .XX X..
    0x83b, // 04073  X.. ... XXX .XX
    0x707, // 03407 .XX X.. ... XXX
  ];

  final BitMatrix _image;

  late bool _compact;
  late int _nbLayers;
  late int _nbDataBlocks;
  late int _nbCenterLayers;
  late int _shift;

  Detector(this._image);

  /// Detects an Aztec Code in an image.
  ///
  /// @param isMirror if true, image is a mirror-image of original
  /// @return {@link AztecDetectorResult} encapsulating results of detecting an Aztec Code
  /// @throws NotFoundException if no Aztec Code can be found
  AztecDetectorResult detect([bool isMirror = false]) {
    // 1. Get the center of the aztec matrix
    Point pCenter = _getMatrixCenter();

    // 2. Get the center points of the four diagonal points just outside the bull's eye
    //  [topRight, bottomRight, bottomLeft, topLeft]
    List<ResultPoint> bullsEyeCorners = _getBullsEyeCorners(pCenter);

    if (isMirror) {
      ResultPoint temp = bullsEyeCorners[0];
      bullsEyeCorners[0] = bullsEyeCorners[2];
      bullsEyeCorners[2] = temp;
    }

    // 3. Get the size of the matrix and other parameters from the bull's eye
    _extractParameters(bullsEyeCorners);

    // 4. Sample the grid
    BitMatrix bits = _sampleGrid(
        _image,
        bullsEyeCorners[_shift % 4],
        bullsEyeCorners[(_shift + 1) % 4],
        bullsEyeCorners[(_shift + 2) % 4],
        bullsEyeCorners[(_shift + 3) % 4]);

    // 5. Get the corners of the matrix.
    List<ResultPoint> corners = _getMatrixCornerPoints(bullsEyeCorners);

    return AztecDetectorResult(
        bits, corners, _compact, _nbDataBlocks, _nbLayers);
  }

  /// Extracts the number of data layers and data blocks from the layer around the bull's eye.
  ///
  /// @param bullsEyeCorners the array of bull's eye corners
  /// @throws NotFoundException in case of too many errors or invalid parameters
  void _extractParameters(List<ResultPoint> bullsEyeCorners) {
    if (!_isValidPoint(bullsEyeCorners[0]) ||
        !_isValidPoint(bullsEyeCorners[1]) ||
        !_isValidPoint(bullsEyeCorners[2]) ||
        !_isValidPoint(bullsEyeCorners[3])) {
      throw NotFoundException.instance;
    }
    int length = 2 * _nbCenterLayers;
    // Get the bits around the bull's eye
    List<int> sides = [
      _sampleLine(bullsEyeCorners[0], bullsEyeCorners[1], length), // Right side
      _sampleLine(bullsEyeCorners[1], bullsEyeCorners[2], length), // Bottom
      _sampleLine(bullsEyeCorners[2], bullsEyeCorners[3], length), // Left side
      _sampleLine(bullsEyeCorners[3], bullsEyeCorners[0], length) // Top
    ];

    // bullsEyeCorners[shift] is the corner of the bulls'eye that has three
    // orientation marks.
    // sides[shift] is the row/column that goes from the corner with three
    // orientation marks to the corner with two.
    _shift = _getRotation(sides, length);

    // Flatten the parameter bits into a single 28- or 40-bit long
    int parameterData = 0;
    for (int i = 0; i < 4; i++) {
      int side = sides[(_shift + i) % 4];
      if (_compact) {
        // Each side of the form ..XXXXXXX. where Xs are parameter data
        parameterData <<= 7;
        parameterData += (side >> 1) & 0x7F;
      } else {
        // Each side of the form ..XXXXX.XXXXX. where Xs are parameter data
        parameterData <<= 10;
        parameterData += ((side >> 2) & (0x1f << 5)) + ((side >> 1) & 0x1F);
      }
    }

    // Corrects parameter data using RS.  Returns just the data portion
    // without the error correction.
    int correctedData = _getCorrectedParameterData(parameterData, _compact);

    if (_compact) {
      // 8 bits:  2 bits layers and 6 bits data blocks
      _nbLayers = (correctedData >> 6) + 1;
      _nbDataBlocks = (correctedData & 0x3F) + 1;
    } else {
      // 16 bits:  5 bits layers and 11 bits data blocks
      _nbLayers = (correctedData >> 11) + 1;
      _nbDataBlocks = (correctedData & 0x7FF) + 1;
    }
  }

  static int _getRotation(List<int> sides, int length) {
    // In a normal pattern, we expect to See
    //   **    .*             D       A
    //   *      *
    //
    //   .      *
    //   ..    ..             C       B
    //
    // Grab the 3 bits from each of the sides the form the locator pattern and concatenate
    // into a 12-bit integer.  Start with the bit at A
    int cornerBits = 0;
    for (int side in sides) {
      // XX......X where X's are orientation marks
      int t = ((side >> (length - 2)) << 1) + (side & 1);
      cornerBits = (cornerBits << 3) + t;
    }
    // Mov the bottom bit to the top, so that the three bits of the locator pattern at A are
    // together.  cornerBits is now:
    //  3 orientation bits at A || 3 orientation bits at B || ... || 3 orientation bits at D
    cornerBits = ((cornerBits & 1) << 11) + (cornerBits >> 1);
    // The result shift indicates which element of List<BullsEyeCorners> goes into the top-left
    // corner. Since the four rotation values have a Hamming distance of 8, we
    // can easily tolerate two errors.
    for (int shift = 0; shift < 4; shift++) {
      if (MathUtils.bitCount(cornerBits ^ _EXPECTED_CORNER_BITS[shift]) <= 2) {
        return shift;
      }
    }
    throw NotFoundException.instance;
  }

  /// Corrects the parameter bits using Reed-Solomon algorithm.
  ///
  /// @param parameterData parameter bits
  /// @param compact true if this is a compact Aztec code
  /// @throws NotFoundException if the array contains too many errors
  static int _getCorrectedParameterData(int parameterData, bool compact) {
    int numCodewords;
    int numDataCodewords;

    if (compact) {
      numCodewords = 7;
      numDataCodewords = 2;
    } else {
      numCodewords = 10;
      numDataCodewords = 4;
    }

    int numECCodewords = numCodewords - numDataCodewords;
    Int32List parameterWords = Int32List(numCodewords);
    for (int i = numCodewords - 1; i >= 0; --i) {
      parameterWords[i] = parameterData & 0xF;
      parameterData >>= 4;
    }
    try {
      ReedSolomonDecoder rsDecoder =
          ReedSolomonDecoder(GenericGF.aztecParam);
      rsDecoder.decode(parameterWords, numECCodewords);
    } catch (ignored) {
      // ReedSolomonException
      throw NotFoundException.instance;
    }
    // Toss the error correction.  Just return the data as an integer
    int result = 0;
    for (int i = 0; i < numDataCodewords; i++) {
      result = (result << 4) + parameterWords[i];
    }
    return result;
  }

  /// Finds the corners of a bull-eye centered on the passed point.
  /// This returns the centers of the diagonal points just outside the bull's eye
  /// Returns [topRight, bottomRight, bottomLeft, topLeft]
  ///
  /// @param pCenter Center point
  /// @return The corners of the bull-eye
  /// @throws NotFoundException If no valid bull-eye can be found
  List<ResultPoint> _getBullsEyeCorners(Point pCenter) {
    Point pina = pCenter;
    Point pinb = pCenter;
    Point pinc = pCenter;
    Point pind = pCenter;

    bool color = true;

    for (_nbCenterLayers = 1; _nbCenterLayers < 9; _nbCenterLayers++) {
      Point pouta = _getFirstDifferent(pina, color, 1, -1);
      Point poutb = _getFirstDifferent(pinb, color, 1, 1);
      Point poutc = _getFirstDifferent(pinc, color, -1, 1);
      Point poutd = _getFirstDifferent(pind, color, -1, -1);

      //d      a
      //
      //c      b

      if (_nbCenterLayers > 2) {
        double q = _distance(poutd, pouta) *
            _nbCenterLayers /
            (_distance(pind, pina) * (_nbCenterLayers + 2));
        if (q < 0.75 ||
            q > 1.25 ||
            !_isWhiteOrBlackRectangle(pouta, poutb, poutc, poutd)) {
          break;
        }
      }

      pina = pouta;
      pinb = poutb;
      pinc = poutc;
      pind = poutd;

      color = !color;
    }

    if (_nbCenterLayers != 5 && _nbCenterLayers != 7) {
      throw NotFoundException.instance;
    }

    _compact = _nbCenterLayers == 5;

    // Expand the square by .5 pixel in each direction so that we're on the border
    // between the white square and the black square
    ResultPoint pinax = ResultPoint(pina.getX() + 0.5, pina.getY() - 0.5);
    ResultPoint pinbx = ResultPoint(pinb.getX() + 0.5, pinb.getY() + 0.5);
    ResultPoint pincx = ResultPoint(pinc.getX() - 0.5, pinc.getY() + 0.5);
    ResultPoint pindx = ResultPoint(pind.getX() - 0.5, pind.getY() - 0.5);

    // Expand the square so that its corners are the centers of the points
    // just outside the bull's eye.
    return _expandSquare([pinax, pinbx, pincx, pindx], 2 * _nbCenterLayers - 3,
        2 * _nbCenterLayers);
  }

  /// Finds a candidate center point of an Aztec code from an image
  ///
  /// @return the center point
  Point _getMatrixCenter() {
    ResultPoint pointA;
    ResultPoint pointB;
    ResultPoint pointC;
    ResultPoint pointD;

    //Get a white rectangle that can be the border of the matrix in center bull's eye or
    try {
      List<ResultPoint> cornerPoints =
          WhiteRectangleDetector(_image).detect();
      pointA = cornerPoints[0];
      pointB = cornerPoints[1];
      pointC = cornerPoints[2];
      pointD = cornerPoints[3];
    } catch (e) {
      // NotFoundException

      // This exception can be in case the initial rectangle is white
      // In that case, surely in the bull's eye, we try to expand the rectangle.
      int cx = _image.getWidth() ~/ 2;
      int cy = _image.getHeight() ~/ 2;
      pointA = _getFirstDifferent(Point(cx + 7, cy - 7), false, 1, -1)
          .toResultPoint();
      pointB = _getFirstDifferent(Point(cx + 7, cy + 7), false, 1, 1)
          .toResultPoint();
      pointC = _getFirstDifferent(Point(cx - 7, cy + 7), false, -1, 1)
          .toResultPoint();
      pointD = _getFirstDifferent(Point(cx - 7, cy - 7), false, -1, -1)
          .toResultPoint();
    }

    //Compute the center of the rectangle
    int cx = MathUtils.round(
        (pointA.getX() + pointD.getX() + pointB.getX() + pointC.getX()) / 4.0);
    int cy = MathUtils.round(
        (pointA.getY() + pointD.getY() + pointB.getY() + pointC.getY()) / 4.0);

    // Redetermine the white rectangle starting from previously computed center.
    // This will ensure that we end up with a white rectangle in center bull's eye
    // in order to compute a more accurate center.
    try {
      List<ResultPoint> cornerPoints =
          WhiteRectangleDetector(_image, 15, cx, cy).detect();
      pointA = cornerPoints[0];
      pointB = cornerPoints[1];
      pointC = cornerPoints[2];
      pointD = cornerPoints[3];
    } catch (e) {
      // NotFoundException
      // This exception can be in case the initial rectangle is white
      // In that case we try to expand the rectangle.
      pointA = _getFirstDifferent(Point(cx + 7, cy - 7), false, 1, -1)
          .toResultPoint();
      pointB = _getFirstDifferent(Point(cx + 7, cy + 7), false, 1, 1)
          .toResultPoint();
      pointC = _getFirstDifferent(Point(cx - 7, cy + 7), false, -1, 1)
          .toResultPoint();
      pointD = _getFirstDifferent(Point(cx - 7, cy - 7), false, -1, -1)
          .toResultPoint();
    }

    // Recompute the center of the rectangle
    cx = MathUtils.round(
        (pointA.getX() + pointD.getX() + pointB.getX() + pointC.getX()) / 4.0);
    cy = MathUtils.round(
        (pointA.getY() + pointD.getY() + pointB.getY() + pointC.getY()) / 4.0);

    return Point(cx, cy);
  }

  /// Gets the Aztec code corners from the bull's eye corners and the parameters.
  ///
  /// @param bullsEyeCorners the array of bull's eye corners
  /// @return the array of aztec code corners
  List<ResultPoint> _getMatrixCornerPoints(List<ResultPoint> bullsEyeCorners) {
    return _expandSquare(bullsEyeCorners, 2 * _nbCenterLayers, _getDimension());
  }

  /// Creates a BitMatrix by sampling the provided image.
  /// topLeft, topRight, bottomRight, and bottomLeft are the centers of the squares on the
  /// diagonal just outside the bull's eye.
  BitMatrix _sampleGrid(BitMatrix image, ResultPoint topLeft,
      ResultPoint topRight, ResultPoint bottomRight, ResultPoint bottomLeft) {
    GridSampler sampler = GridSampler.getInstance();
    int dimension = _getDimension();

    double low = dimension / 2.0 - _nbCenterLayers;
    double high = dimension / 2.0 + _nbCenterLayers;

    return sampler.sampleGridBulk(
        image,
        dimension,
        dimension,
        low,
        low, // topleft
        high,
        low, // topright
        high,
        high, // bottomright
        low,
        high, // bottomleft
        topLeft.getX(),
        topLeft.getY(),
        topRight.getX(),
        topRight.getY(),
        bottomRight.getX(),
        bottomRight.getY(),
        bottomLeft.getX(),
        bottomLeft.getY());
  }

  /// Samples a line.
  ///
  /// @param p1   start point (inclusive)
  /// @param p2   end point (exclusive)
  /// @param size number of bits
  /// @return the array of bits as an int (first bit is high-order bit of result)
  int _sampleLine(ResultPoint p1, ResultPoint p2, int size) {
    int result = 0;

    double d = _distanceResult(p1, p2);
    double moduleSize = d / size;
    double px = p1.getX();
    double py = p1.getY();
    double dx = moduleSize * (p2.getX() - p1.getX()) / d;
    double dy = moduleSize * (p2.getY() - p1.getY()) / d;
    for (int i = 0; i < size; i++) {
      if (_image.get(
          MathUtils.round(px + i * dx), MathUtils.round(py + i * dy))) {
        result |= 1 << (size - i - 1);
      }
    }
    return result;
  }

  /// @return true if the border of the rectangle passed in parameter is compound of white points only
  ///         or black points only
  bool _isWhiteOrBlackRectangle(Point p1, Point p2, Point p3, Point p4) {
    int corr = 3;

    p1 = Point(p1.getX() - corr, p1.getY() + corr);
    p2 = Point(p2.getX() - corr, p2.getY() - corr);
    p3 = Point(p3.getX() + corr, p3.getY() - corr);
    p4 = Point(p4.getX() + corr, p4.getY() + corr);

    int cInit = _getColor(p4, p1);

    if (cInit == 0) {
      return false;
    }

    int c = _getColor(p1, p2);

    if (c != cInit) {
      return false;
    }

    c = _getColor(p2, p3);

    if (c != cInit) {
      return false;
    }

    c = _getColor(p3, p4);

    return c == cInit;
  }

  /// Gets the color of a segment
  ///
  /// @return 1 if segment more than 90% black, -1 if segment is more than 90% white, 0 else
  int _getColor(Point p1, Point p2) {
    double d = _distance(p1, p2);
    double dx = (p2.getX() - p1.getX()) / d;
    double dy = (p2.getY() - p1.getY()) / d;
    int error = 0;

    double px = p1.getX().toDouble();
    double py = p1.getY().toDouble();

    bool colorModel = _image.get(p1.getX(), p1.getY());

    int iMax = d.ceil();
    for (int i = 0; i < iMax; i++) {
      px += dx;
      py += dy;
      if (_image.get(MathUtils.round(px), MathUtils.round(py)) != colorModel) {
        error++;
      }
    }

    double errRatio = error / d;

    if (errRatio > 0.1 && errRatio < 0.9) {
      return 0;
    }

    return (errRatio <= 0.1) == colorModel ? 1 : -1;
  }

  /// Gets the coordinate of the first point with a different color in the given direction
  Point _getFirstDifferent(Point init, bool color, int dx, int dy) {
    int x = init.getX() + dx;
    int y = init.getY() + dy;

    while (_isValid(x, y) && _image.get(x, y) == color) {
      x += dx;
      y += dy;
    }

    x -= dx;
    y -= dy;

    while (_isValid(x, y) && _image.get(x, y) == color) {
      x += dx;
    }
    x -= dx;

    while (_isValid(x, y) && _image.get(x, y) == color) {
      y += dy;
    }
    y -= dy;

    return Point(x, y);
  }

  /// Expand the square represented by the corner points by pushing out equally in all directions
  ///
  /// @param cornerPoints the corners of the square, which has the bull's eye at its center
  /// @param oldSide the original length of the side of the square in the target bit matrix
  /// @param newSide the new length of the size of the square in the target bit matrix
  /// @return the corners of the expanded square
  static List<ResultPoint> _expandSquare(
      List<ResultPoint> cornerPoints, int oldSide, int newSide) {
    double ratio = newSide / (2.0 * oldSide);
    double dx = cornerPoints[0].getX() - cornerPoints[2].getX();
    double dy = cornerPoints[0].getY() - cornerPoints[2].getY();
    double centerx = (cornerPoints[0].getX() + cornerPoints[2].getX()) / 2.0;
    double centery = (cornerPoints[0].getY() + cornerPoints[2].getY()) / 2.0;

    ResultPoint result0 =
        ResultPoint(centerx + ratio * dx, centery + ratio * dy);
    ResultPoint result2 =
        ResultPoint(centerx - ratio * dx, centery - ratio * dy);

    dx = cornerPoints[1].getX() - cornerPoints[3].getX();
    dy = cornerPoints[1].getY() - cornerPoints[3].getY();
    centerx = (cornerPoints[1].getX() + cornerPoints[3].getX()) / 2.0;
    centery = (cornerPoints[1].getY() + cornerPoints[3].getY()) / 2.0;
    ResultPoint result1 =
        ResultPoint(centerx + ratio * dx, centery + ratio * dy);
    ResultPoint result3 =
        ResultPoint(centerx - ratio * dx, centery - ratio * dy);

    return [result0, result1, result2, result3];
  }

  bool _isValid(int x, int y) {
    return x >= 0 && x < _image.getWidth() && y > 0 && y < _image.getHeight();
  }

  bool _isValidPoint(ResultPoint point) {
    int x = MathUtils.round(point.getX());
    int y = MathUtils.round(point.getY());
    return _isValid(x, y);
  }

  static double _distance(Point a, Point b) {
    return MathUtils.distance(a.getX().toDouble(), a.getY().toDouble(),
        b.getX().toDouble(), b.getY().toDouble());
  }

  static double _distanceResult(ResultPoint a, ResultPoint b) {
    return MathUtils.distance(a.getX(), a.getY(), b.getX(), b.getY());
  }

  int _getDimension() {
    if (_compact) {
      return 4 * _nbLayers + 11;
    }
    return 4 * _nbLayers + 2 * ((2 * _nbLayers + 6) ~/ 15) + 15;
  }
}
